# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

import os
import unittest

import numpy as np

import thermofeel as tmf


def data_file(name):
    return os.path.join(os.path.dirname(__file__), name)


# combining the pytest library with the functionality of np testing for use with np.array file type


class TestThermalCalculator(unittest.TestCase):
    def setUp(self):
        # variables for use in thermalindexcalculator
        t = np.genfromtxt(
            data_file("thermofeel_testcases.csv"),
            delimiter=",",
            names=True,
        )
        self.t2m = t["t2m"]
        self.ssr = t["ssr"]
        self.td = t["td"]
        self.va = t["va"]
        self.mrt = t["mrt"]
        self.ssrd = t["ssrd"]
        self.strd = t["strd"]
        self.fdir = t["fdir"]
        self.strr = t["strr"]
        self.cossza = t["cossza"]
        self.phase = t["phase"]
        self.va_height = t["va_height"]

        self.rh = np.loadtxt(data_file("rh.csv"))
        self.es = np.loadtxt(data_file("es.csv"))
        self.ens = np.loadtxt(data_file("ens.csv"))
        self.es_multiphase = np.loadtxt(data_file("es_multiphase.csv"))
        self.va_scaled = np.loadtxt(data_file("va_scaled.csv"))
        self.utci = np.loadtxt(data_file("utci.csv"))
        self.wbgts = np.loadtxt(data_file("wbgts.csv"))
        self.wbt = np.loadtxt(data_file("wbt.csv"))
        self.bgt = np.loadtxt(data_file("bgt.csv"))
        self.wbgt = np.loadtxt(data_file("wbgt.csv"))
        self.humidex = np.loadtxt(data_file("humidex.csv"))
        self.net = np.loadtxt(data_file("net.csv"))
        self.at = np.loadtxt(data_file("at.csv"))
        self.windchill = np.loadtxt(data_file("windchill.csv"))
        self.heatindex = np.loadtxt(data_file("heatindex.csv"))
        self.heatindexadjusted = np.loadtxt(data_file("hia.csv"))
        self.mrtr = np.loadtxt(data_file("mrtr.csv"))
        self.mrt_from_bgt = np.loadtxt(data_file("mrt_from_bgt.csv"))

        self.dsrp = tmf.approximate_dsrp(self.fdir, self.cossza)

    def assert_equal(self, expected, result, decimal=6):
        self.assertequal = self.assertIsNone(
            np.testing.assert_array_almost_equal(expected, result, decimal)
        )

    def test_relative_humidity_percent(self):
        rh = tmf.calculate_relative_humidity_percent(self.t2m, self.td)
        # np.savetxt("rh.csv", rh)
        self.assert_equal(self.rh, rh)

    def test_saturation_vapour_pressure(self):
        es = tmf.calculate_saturation_vapour_pressure(self.t2m)
        # np.savetxt("es.csv", es)
        self.assert_equal(self.es, es)

    def test_saturation_vapour_pressure_multiphase(self):
        es_multiphase = tmf.calculate_saturation_vapour_pressure_multiphase(
            self.t2m, self.phase
        )
        # np.savetxt("es_multiphase.csv", es_multiphase)
        self.assert_equal(self.es_multiphase, es_multiphase)

    def test_nonsaturation_vapour_pressure(self):
        rh = tmf.calculate_relative_humidity_percent(self.t2m, self.td)
        ens = tmf.calculate_nonsaturation_vapour_pressure(self.t2m, rh)
        # np.savetxt("ens.csv", ens)
        self.assert_equal(self.ens, ens)

    def test_scale_windspeed(self):
        va_scaled = tmf.scale_windspeed(self.va, self.va_height)
        # np.savetxt("va_scaled.csv", va_scaled)
        self.assert_equal(self.va_scaled, va_scaled)

    def test_dew_point_from_relative_humidity(self):
        td = tmf.calculate_dew_point_from_relative_humidity(self.rh, self.t2m)
        self.assert_equal(self.td, td, decimal=2)

    def test_mean_radiant_temperature(self):
        mrtr = tmf.calculate_mean_radiant_temperature(
            ssrd=self.ssrd / 3600,
            ssr=self.ssr / 3600,
            dsrp=self.dsrp / 3600,
            strd=self.strd / 3600,
            fdir=self.fdir / 3600,
            strr=self.strr / 3600,
            cossza=self.cossza / 3600,
        )
        # np.savetxt("mrtr.csv", mrtr)
        self.assert_equal(self.mrtr, mrtr)  # , decimal = 3)

    def test_utci(self):
        rh_pc = tmf.calculate_relative_humidity_percent(self.t2m, self.td)
        ehPa = tmf.calculate_saturation_vapour_pressure(self.t2m) * rh_pc / 100.0
        utci = tmf.calculate_utci(
            t2_k=self.t2m, va=self.va, mrt=self.mrt, td_k=None, ehPa=ehPa
        )
        # np.savetxt("utci.csv", utci)
        self.assert_equal(self.utci, utci)

    def test_wbgt_simple(self):
        rh_pc = tmf.calculate_relative_humidity_percent(self.t2m, self.td)
        wbgts = tmf.calculate_wbgt_simple(self.t2m, rh_pc)
        # np.savetxt("wbgts.csv", wbgts)
        self.assert_equal(self.wbgts, wbgts)

    def test_wbt(self):
        rh_pc = tmf.calculate_relative_humidity_percent(self.t2m, self.td)
        wbt = tmf.calculate_wbt(self.t2m, rh_pc)
        # np.savetxt("wbt.csv", wbt)
        self.assert_equal(self.wbt, wbt)

    def test_bgt(self):
        bgt = tmf.calculate_bgt(self.t2m, self.mrt, self.va)
        # np.savetxt("bgt.csv", bgt)
        self.assert_equal(self.bgt, bgt)

    def test_wbgt(self):
        wbgt = tmf.calculate_wbgt(self.t2m, self.mrt, self.va, self.td)
        # np.savetxt("wbgt.csv", wbgt)
        self.assert_equal(self.wbgt, wbgt)

    def test_mrt_from_bgt(self):
        mrt_from_bgt = tmf.calculate_mrt_from_bgt(self.t2m, self.bgt, self.va)
        # np.savetxt("mrt_from_bgt.csv", mrt_from_bgt)
        self.assert_equal(self.mrt_from_bgt, mrt_from_bgt)

    def test_humidex(self):
        humidex = tmf.calculate_humidex(self.t2m, self.td)
        # np.savetxt("humidex.csv", humidex)
        self.assert_equal(self.humidex, humidex)

    def test_normal_effective_temperature(self):
        rh_pc = tmf.calculate_relative_humidity_percent(self.t2m, self.td)
        net = tmf.calculate_normal_effective_temperature(self.t2m, self.va, rh_pc)
        # np.savetxt("net.csv", net)
        self.assert_equal(self.net, net)

    def test_apparent_temperature(self):
        rh_pc = tmf.calculate_relative_humidity_percent(self.t2m, self.td)
        at = tmf.calculate_apparent_temperature(self.t2m, self.va, rh_pc)
        # np.savetxt("at.csv", at)
        self.assert_equal(self.at, at)

    def test_wind_chill(self):
        windchill = tmf.calculate_wind_chill(self.t2m, self.va)
        # np.savetxt("windchill.csv", windchill)
        self.assert_equal(self.windchill, windchill)

    def test_heat_index(self):
        rh_pc = tmf.calculate_relative_humidity_percent(self.t2m, self.td)
        heatindex = tmf.calculate_heat_index_simplified(self.t2m, rh_pc)
        # np.savetxt("heatindex.csv", heatindex)
        self.assert_equal(self.heatindex, heatindex)

    def test_heat_index_adjusted(self):
        hia = tmf.calculate_heat_index_adjusted(self.t2m, self.td)
        # np.savetxt("hia.csv", hia)
        self.assert_equal(self.heatindexadjusted, hia)


if __name__ == "__main__":
    unittest.main()  # pragma: no cover



import math

import cartopy.crs as ccrs
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from netCDF4 import Dataset, date2num, num2date

from thermofeel import (
    calculate_cos_solar_zenith_angle,
    calculate_relative_humidity_percent,
    calculate_saturation_vapour_pressure_multiphase,
)

# from datetime import datetime


to_radians = math.pi / 180.0

# physical constants
global M_AIR, M_H2O, R_GAS, Cp, STEFANB, RATIO, R_AIR, Pr
M_AIR = 28.97  # molecular weight of dry air (grams per mole)
M_H2O = 18.015  # molecular weight of water vapor (grams per mole)
R_GAS = 8314.34  # ideal gas constant (J/kg mol · K)
Cp = 1003.5  # Specific heat capacity of air at constant pressure (J·kg-1·K-1)
STEFANB = 0.000000056696  # stefan-boltzmann constant
RATIO = Cp * M_AIR / M_H2O
R_AIR = R_GAS * (M_AIR ** (-1))
Pr = Cp * ((Cp + 1.25 * R_AIR) ** (-1))  # Prandtl number

# globe constants
global D_GLOBE, EMIS_GLOBE, ALB_GLOBE
D_GLOBE = 0.0508  # diameter of globe (m)
EMIS_GLOBE = 0.95  # emissivity of globe
ALB_GLOBE = 0.05  # albedo of globe

# wick constants
global EMIS_WICK, ALB_WICK, D_WICK, L_WICK
EMIS_WICK = 0.95  # emissivity of the wick
ALB_WICK = 0.4  # albedo of the wick
D_WICK = 0.007  # diameter of the wick
L_WICK = 0.0254  # length of the wick

# surface constant
global EMIS_SFC, ALB_SFC
EMIS_SFC = 0.999
ALB_SFC = 0.45

global MAX_ITER, CONVERGENCE
MAX_ITER = 50
CONVERGENCE = 0.02

global MISSING_VALUE
MISSING_VALUE = -9999.0
MIN_SPEED = 0.13


def viscosity(t2m):
    """
    Calculate viscosity
    :param t2m: (float array) 2m temperature [K]
    returns  air viscosity [kg/(m/s)]

    """
    omega = 1.2945 - t2m / 1141.176470588
    visc = 0.0000026693 * (np.sqrt(28.97 * t2m)) * ((13.082689 * omega) ** (-1))
    return visc


def thermal_cond(tk):
    """
    Calculate thermal conductivity
    :param tk: (float array) temperature [K]
    :param cp: (int) Specific heat capacity of air at constant pressure (J·kg-1·K-1)
    :param R_AIR: (int) physical constant
    returns  air viscosity [kg/(m/s)]
    """
    tc = (Cp + 1.25 * R_AIR) * viscosity(tk)
    return tc


def emis_atm(t2m, rhpc):
    """
    Purpose: calculate the atmospheric emissivity.
    Reference: Oke (2nd edition), page 373.
    :param t2m: 2m temperature [K]
    :param rh: relative humidity must be in % [0-100]
    :return:
    """
    liquid = 0
    esat = calculate_saturation_vapour_pressure_multiphase(t2m, liquid)

    rh = rhpc * 0.01
    e = rh * esat
    emis_atm = 0.575 * (e**0.143)
    return emis_atm


def diffusivity(tk, ps):
    # tk: air temperature (K)
    # ps: surface pressure (Pa)
    # return diffusivity of water vapor in air (m2/s)

    diff = (
        2.471773765165648e-05
        * ((tk * 0.0034210563748421257) ** 2.334)
        * ((ps / 101325) ** (-1))
    )
    return diff


def evap(tk):
    # tk: air temperature (K)
    # return heat of evaporation (J/(kg K))
    hevap = 1665134.5 + 2370.0 * tk

    return hevap


def h_sphere_and_cylinder_in_air(t2m, ps, va, diamglobe, diamwick, Pr, cp, R_AIR):
    # tas: air temperature (K)
    # ps: surface pressure (Pa)
    # sfcwind: 2 meter wind (m/s)
    # return convective heat tranfer coefficient for flow around a sphere (W/(m2 K))
    # h = h_sphere_and_cylinder_in_air(tref,ps,va,diamglobe,diamwick,Pr,cp,R_AIR)
    thermcon = thermal_cond(t2m)
    density = ps * ((R_AIR * t2m) ** (-1))
    Re_globe = va * density * diamglobe * ((viscosity(t2m)) ** (-1))
    Nu_globe = 2 + 0.6 * np.sqrt(Re_globe) * np.power(Pr, 0.3333)
    h_globe = Nu_globe * thermcon * (diamglobe ** (-1))

    Re_wick = va * density * diamwick * ((viscosity(t2m)) ** (-1))
    Nu_wick = 0.281 * (Re_wick**0.6) * (Pr**0.44)
    h_wick = Nu_wick * thermcon * (diamwick ** (-1))

    return h_globe, h_wick


def h_sphere_in_air(diameter, Tair, Pair, speed):
    """
    Calculate the convective heat transfer coefficient, W/(m2 K), for flow around a sphere.
    diameter, sphere diameter, m
        Tair, air temperature, K
        Pair, barometric pressure, mb
        speed, fluid (air) speed, m/s
    Reference: Bird, Stewart, and Lightfoot (BSL), page 409.
    """
    density = Pair * 100.0 / (R_AIR * Tair)
    Re = max(speed, MIN_SPEED) * density * diameter / viscosity(Tair)
    Nu = 2.0 + 0.6 * math.sqrt(Re) * pow(Pr, 0.3333)
    return Nu * thermal_cond(Tair) / diameter


def h_cylinder_in_air(diameter, length, Tair, Pair, speed):
    """
    diameter, cylinder diameter, [m]
        length,   cylinder length, [m] - does not seem to be used
        Tair, air temperature, [K]
        Pair, barometric pressure, [hPa]
        speed,	fluid (wind) speed, [m/s]
    """
    # parameters from Bedingfield and Drew
    a = 0.56
    b = 0.281
    c = 0.4
    density = (Pair * 100.0) / (R_AIR * Tair)
    Re = max(speed, MIN_SPEED) * density * diameter / viscosity(Tair)
    Nu = b * pow(Re, (1.0 - c)) * pow(Pr, (1.0 - a))
    return Nu * thermal_cond(Tair) / diameter


def bgt_lijigren(t2m, rh, ps, ssrd, fdir, cossza, va):
    """
    Purpose: to calculate the globe temperature.
    Method by J. Liljegren, Argonne National Laboratory
    """
    Tair = t2m
    speed = va
    Pair = ps
    cza = cossza
    solar = ssrd

    Tsfc = Tair
    Tglobe_prev = Tair
    # first guess is the air temperature

    converged = False
    iter = 0
    while not converged and iter < MAX_ITER:
        iter = iter + 1
        Tref = 0.5 * (
            Tglobe_prev + Tair
        )  # evaluate properties at the average temperature
        h = h_sphere_in_air(D_GLOBE, Tref, Pair, speed)
        Tglobe_new = pow(
            0.5 * (emis_atm(Tair, rh) * pow(Tair, 4.0) + EMIS_SFC * pow(Tsfc, 4.0))
            - h / (STEFANB * EMIS_GLOBE) * (Tglobe_prev - Tair)
            + solar
            / (2.0 * STEFANB * EMIS_GLOBE)
            * (1.0 - ALB_GLOBE)
            * (fdir * (1.0 / (2.0 * cza) - 1.0) + 1.0 + ALB_SFC),
            0.25,
        )

        if abs(Tglobe_new - Tglobe_prev) < CONVERGENCE:
            converged = True

        Tglobe_prev = 0.9 * Tglobe_prev + 0.1 * Tglobe_new

    if converged:
        return Tglobe_new - 273.15
    else:
        return MISSING_VALUE

    # bgt = bgt_lijigren(t2m, rh, ps, ssrd/3600, fdir_frac, cossza, eatm, diamglobe, diamwick, Pr, va, cp, R_AIR)
    #
    # equation of Tg that needs to be solved by iteration
    # i = 0
    # diamglobe = diamglobe
    # diamwick= diamwick
    # Pr = Pr
    # va = va
    # tsfc = t2m
    # albsfc = 0.45
    # albglobe = 0.05
    # eglobe = 0.95
    # tglobe_prev = np.copy(t2m) #first guess is 2m temperature
    # tref = 0.5 * (tglobe_prev + t2m)
    # esfc = 0.999
    # h = h_sphere_and_cylinder_in_air(tref,ps,va,diamglobe,diamwick,Pr,cp,R_AIR)
    # h = h[0]
    # # while i < 50:
    # #     tg = np.power(0.5*( eatm * np.power(t2m,4)) + esfc * np.power(tsfc,4) \
    # #                 - h/(STEFANB*eglobe)*(tglobe_prev - t2m) \
    # #                     + ssrd/(2.*STEFANB*eglobe)*(1.-albglobe)\
    # #                         *(fdir*(1./(2.*cossza)-1.)+1.+ albsfc), 0.25)
    # #     tg_filter = np.where(np.abs(tg - tglobe_prev) < tg)
    # #     tglobe_prev[tg_filter] = 0.9 * tglobe_prev[tg_filter] + 0.1* tg[tg_filter]
    # #     print(np.average(tg))
    # #     i = i + 1
    # # tg = np.power(0.5*(eatm * np.power(t2m, 4)) + esfc * np.power(tsfc, 4) - h/(STEFANB*eglobe)*(tglobe_prev - t2m)
    #      + ssrd/(2.*STEFANB*eglobe)*(1.-albglobe) * (fdir*(1./(2.*cossza)-1.)+1.+ albsfc), 0.25)
    # tg = np.power(0.5 * ((1+eatm) * np.power(t2m, 4)) - h/(STEFANB*eglobe)*(tglobe_prev - t2m)
    #    + ssrd/(2.*STEFANB*eglobe)*(1.-albglobe) * (fdir*(1./(2.*cossza)-1.)+1.+ albsfc), 0.25)
    # tg_filter = np.where(np.abs(tg - tglobe_prev) < tg)
    # tglobe_prev[tg_filter] = 0.9 * tglobe_prev[tg_filter] + 0.1* tg[tg_filter]
    # print(np.average(tg))
    # return tg


def wbt_lijigren(t2m, td, rh, ps, va, ssrd, fdir, cossza):
    """
    Purpose: to calculate the natural wet bulb temperature.
    Method by J. Liljegren, Argonne National Laboratory
    """

    Tair = t2m
    Pair = ps
    speed = va
    solar = ssrd

    radiative = 1
    liquid = 0

    a = 0.56  # from Bedingfield and Drew

    Tsfc = Tair
    sza = math.acos(cossza)
    eair = rh * calculate_saturation_vapour_pressure_multiphase(t2m, liquid)
    Tdew = td
    Twb_prev = Tdew  # first guess is the dew point temperature

    converged = False
    iter = 0
    while not converged and iter < MAX_ITER:
        iter = iter + 1
        Tref = 0.5 * (Twb_prev + Tair)  # evaluate properties at the average temperature
        h = h_cylinder_in_air(D_WICK, L_WICK, Tref, Pair, speed)
        Fatm = STEFANB * EMIS_WICK * (
            0.5 * (emis_atm(Tair, rh) * pow(Tair, 4.0) + EMIS_SFC * pow(Tsfc, 4.0))
            - pow(Twb_prev, 4.0)
        ) + (1.0 - ALB_WICK) * solar * (
            (1.0 - fdir) * (1.0 + 0.25 * D_WICK / L_WICK)
            + fdir * ((math.tan(sza) / math.pi) + 0.25 * D_WICK / L_WICK)
            + ALB_SFC
        )
        ewick = calculate_saturation_vapour_pressure_multiphase(Twb_prev, liquid)
        density = Pair * 100.0 / (R_AIR * Tref)

        Sc = viscosity(Tref) / (density * diffusivity(Tref, Pair))  # Schmidt number
        print(f"Sc {Sc}")

        Twb_new = (
            Tair
            - evap(Tref) / RATIO * (ewick - eair) / (Pair - ewick) * pow(Pr / Sc, a)
            + (Fatm / h * radiative)
        )

        print(f"Twb_prev {Twb_prev} Twb_new {Twb_new} Increment {Twb_new-Twb_prev}")

        if abs(Twb_new - Twb_prev) < CONVERGENCE:
            converged = True
        Twb_prev = 0.9 * Twb_prev + 0.1 * Twb_new

    if converged:
        return Twb_new - 273.15
    else:
        return MISSING_VALUE

    # h = h_sphere_and_cylinder_in_air(tref,ps,va,diamglobe,diamwick,Pr,cp,R_AIR)
    # h = h[1]
    # tref = 0.5 * (twb_prev + t2m)
    # emis_atm = emis_atm(t2m,rh,ps)
    # emiswick =	0.95
    # albwick =	0.4
    # dwick = 0.007
    # lwick =	0.0254
    # emisfc = 0.999

    # converged = False

    # while i < 50:
    #     Fatm = STEFANB * emiswick *  ( 0.5*( emis_atm *np.power(t2m,4) + emisfc*np.power(tsfc,4) )
    #                                   - np.power(twb_prev,4) ) + (1.-albwick) * ssrd * \


# 	       ( (1.-fdir)*(1.+0.25*dwick/lwick) + fdir*((np.tan(sza)/np.pi)+0.25*dwick/lwick) + albsfc )
#     ewick = calculate_saturation_vapour_pressure(twb_prev)
#     density = ps * 100 / (R_AIR * tref)
#     Sc = viscosity(tref)/(density* diffusivity(t2m, ps))
#     twb = t2m - evap(t2m) / ratio * (ewick - eair)/(ps - ewick) * np.power(Pr/Sc,a) \
#         + (Fatm/h * -1)
#    # twb_filter = np.where(np.abs(twb - twb_prev) < twb)
#     #twb_prev[twb_filter] = 0.9 * twb_prev[twb_filter] + 0.1* twb[twb_filter]

#     i= i + 1

# return twb


def main():
    filename = "March2013LijiCake.nc"
    print("Loading NetCDF file {filename} --  Go and have a coffee ....")

    fh2 = Dataset(filename, mode="r")
    # print(fh2.variables)
    lons = fh2.variables["longitude"][:]  # degrees
    lats = fh2.variables["latitude"][:]  # degrees
    nctime = fh2.variables["time"]  # get values
    dtime = num2date(nctime[1], nctime.units)
    # print(dtime)
    ssrd = fh2.variables["ssrd"][1]  # Jm-2
    fdir = fh2.variables["fdir"][1]  # Jm-2

    lon_mg, lat_mg = np.meshgrid(lons, lats)
    va = np.sqrt(fh2.variables["u10"][1] ** 2 + fh2.variables["v10"][1] ** 2)  # ms-1
    t2m = fh2.variables["t2m"][1]  # K
    td = fh2.variables["d2m"][1]  # K
    ps = fh2.variables["sp"][1] / 100  # hPa
    # rh = 0.62198 * calculate_saturation_vapour_pressure(t2m) / (ps - calculate_saturation_vapour_pressure(t2m))
    rh = calculate_relative_humidity_percent(t2m, td)  # %
    # print(rh)
    # eatm = emis_atm(t2m, rh)  # unitless
    # print(eatm)
    cossza = calculate_cos_solar_zenith_angle(
        lat=lat_mg, lon=lon_mg, y=dtime.year, m=dtime.month, d=dtime.day, h=dtime.hour
    )  # unitless

    #
    # NOTE: CHECK THIS MAKES SENSE OR HAS IMPACT IN RESULTS
    #
    cossza = np.where(
        cossza == 0, 0.25, cossza
    )  # TODO: Check it tolerance checking is needed instead!

    #
    # NOTE: CHECK THIS MAKES SENSE OR HAS IMPACT IN RESULTS
    #
    fdir_frac = np.where(fdir > 0, fdir / ssrd, 0)  # unitless

    ssrd = ssrd / 3600.0
    bgt = np.zeros(len(cossza))
    wbt = np.zeros(len(cossza))
    for i in range(len(cossza)):
        t2m_ = float(t2m[i][1])
        td_ = float(td[i][1])
        rh_ = float(rh[i][1])
        ps_ = float(ps[i][1])
        ssrd_ = float(ssrd[i][1])
        fdir_frac_ = float(fdir_frac[i][1])
        cossza_ = float(cossza[i][1])
        va_ = float(va[i][1])
        bgt[i] = bgt_lijigren(t2m_, rh_, ps_, ssrd_, fdir_frac_, cossza_, va_)
        wbt[i] = wbt_lijigren(t2m_, td_, rh_, ps_, va_, ssrd_, fdir_frac_, cossza_)
        print(f"i {i} -> bgt {bgt[i]} wbt {wbt[i]}")

    WBGTvar = 0.7 * (wbt - 273.15) + 0.2 * (bgt - 273.15) + 0.1 * (t2m - 273.15)

    # print (WBGTvar)
    # h = h_sphere_and_cylinder_in_air(t2m, ps, va,diamglobe,diamwick,Pr,cp,R_AIR)

    ########################

    # fig = plt.figure()
    # ax = plt.axes(projection=ccrs.AlbersEqualArea())
    # ax.coastlines()
    # filled_c = plt.pcolormesh(lon_mg, lat_mg, (h[0]),
    #                           transform=ccrs.PlateCarree(),
    #                           cmap="RdBu_r")
    # fig.colorbar(filled_c, orientation="horizontal",extend="both")
    # plt.savefig("h0.png")
    # plt.clf()
    #
    # fig = plt.figure()
    # ax = plt.axes(projection=ccrs.AlbersEqualArea())
    # ax.coastlines()
    # filled_c = plt.pcolormesh(lon_mg, lat_mg, (h[1]),
    #                           transform=ccrs.PlateCarree(),
    #                           cmap="RdBu_r")
    # fig.colorbar(filled_c, orientation="horizontal",extend="both")
    # plt.savefig("h1.png")
    # plt.clf()

    fig = plt.figure()
    ax = plt.axes(projection=ccrs.AlbersEqualArea())
    ax.coastlines()
    filled_c = plt.pcolormesh(
        lon_mg,
        lat_mg,
        bgt,
        transform=ccrs.PlateCarree(),
        cmap="RdBu_r",
        vmin=273,
        vmax=350,
    )
    fig.colorbar(filled_c, orientation="horizontal", extend="both")
    plt.savefig("bgt.png")
    plt.clf()

    # fig = plt.figure()
    # ax = plt.axes(projection=ccrs.PlateCarree())
    # ax.coastlines()
    # filled_c = plt.pcolormesh(lon_mg, lat_mg, (wbt),
    #                           transform=ccrs.PlateCarree(),
    #                           cmap="RdBu_r")
    # fig.colorbar(filled_c, orientation="horizontal",extend="both")
    # plt.savefig("wbt.png")
    # plt.clf()
    #
    # fig = plt.figure()
    # ax = plt.axes(projection=ccrs.PlateCarree())
    # ax.coastlines()
    # filled_c = plt.pcolormesh(lon_mg, lat_mg, (WBGTvar),
    #                           transform=ccrs.PlateCarree(),
    #                           cmap="RdBu_r")
    # fig.colorbar(filled_c, orientation="horizontal",extend="both")
    # plt.savefig("wbgtvar.png")
    # plt.clf()
    #
    # fig = plt.figure()
    # ax = plt.axes(projection=ccrs.PlateCarree())
    # ax.coastlines()
    # filled_c = plt.pcolormesh(lon_mg, lat_mg, (t2m),
    #                           transform=ccrs.PlateCarree(),
    #                           cmap="RdBu_r")
    # fig.colorbar(filled_c, orientation="horizontal",extend="both")
    # plt.savefig("t2m.png")
    # plt.clf()

    time = pd.date_range(start="2013-03-01", end="2013-03-31").to_pydatetime().tolist()

    ncout = Dataset("wbgt2013liji.nc", "w", format="NETCDF4")
    ncout.createDimension("latitude", len(lat_mg))
    ncout.createDimension("longitude", len(lon_mg))
    ncout.createDimension("time", len(time))

    ncout_latitude = ncout.createVariable("latitude", "f4", ("latitude",))
    ncout_latitude[:] = lat_mg
    ncout_longitude = ncout.createVariable("longitude", "f4", ("longitude",))
    ncout_longitude[:] = lon_mg

    units = "hours in march 2013"
    ncout_time = ncout.createVariable("time", "i4", ("time",))
    ncout_time[:] = date2num(time, units)
    ncout_time.units = units

    ncout_dist = ncout.createVariable("WBGT", "f4", ("time", "latitude", "longitude"))
    ncout_dist.long_name = "WBGT"
    ncout_dist.units = "C"
    ncout_dist[:] = WBGTvar

    ncout.close()


if __name__ == "__main__":
    main()



# (C) Copyright 1996- ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.

# Note: This script is intended only as example usage of thermofeel library.
#       It is designed to be used with ECMWF forecast data.
#       The function ifs_step_intervals() is used to calculate the time interval based on the forecast step.
#       This is particular to the IFS model and ECMWF's NWP operational system.

import argparse
import math
import sys
from datetime import datetime, timedelta, timezone

import eccodes
import numpy as np

import thermofeel as thermofeel

UTCI_MIN_VALUE = thermofeel.celsius_to_kelvin(-80)
UTCI_MAX_VALUE = thermofeel.celsius_to_kelvin(90)
MISSING_VALUE = -9999.0

###########################################################################################################

lats = None
lons = None

results = {}
misses = {}

###########################################################################################################


def field_stats(name, values):
    if name in misses:
        values[misses[name]] = np.nan

    print(
        f"{name} min {np.nanmin(values)} max {np.nanmax(values)} "
        f"avg {np.nanmean(values)} stddev {np.nanstd(values, dtype=np.float64)} "
        f"missing {np.count_nonzero(np.isnan(values))}"
    )

    if name in misses:
        values[misses[name]] = MISSING_VALUE


###########################################################################################################


def decode_grib(fpath):
    print(f"decoding file {fpath}")

    prev_step = None
    prev_number = None

    msgcount = 0
    messages = {}

    with open(fpath, "rb") as f:
        while True:
            msg = eccodes.codes_any_new_from_file(f)

            if msg is None:  # end of file, stop iterating
                # print(f"yielding {len(messages)} messages")
                yield messages
                for k, m in messages.items():
                    grib = m["grib"]
                    eccodes.codes_release(grib)
                messages = {}
                break

            md = dict()
            msgcount += 1

            # loop metadata key-values
            it = eccodes.codes_keys_iterator_new(msg, "mars")
            while eccodes.codes_keys_iterator_next(it):
                k = eccodes.codes_keys_iterator_get_name(it)
                v = eccodes.codes_get_string(msg, k)
                md[k] = v
            eccodes.codes_keys_iterator_delete(it)

            # change types
            step = int(md["step"])
            number = md.get("number", None)

            # on new step or number, return/yield group of messages accumulated so far
            # and ensure proper cleanup of memory

            stop = (prev_step is not None and step != prev_step) or (
                prev_number is not None and number != prev_number
            )

            if stop:
                # print(f"yielding {len(messages)} messages")
                yield messages
                for k, m in messages.items():
                    grib = m["grib"]
                    eccodes.codes_release(grib)
                messages = {}

            prev_number = number
            prev_step = step

            # print(f"message {msgcount} mars metadata: {md}")

            # aggregate messages on step, number, assuming they are contiguous

            md["paramId"] = eccodes.codes_get_string(msg, "paramId")
            md["shortName"] = eccodes.codes_get_string(msg, "shortName")

            md["Ni"] = eccodes.codes_get_long(msg, "Ni")
            md["Nj"] = eccodes.codes_get_long(msg, "Nj")

            md["time"] = eccodes.codes_get_long(msg, "time")
            md["date"] = eccodes.codes_get_string(msg, "date")
            md["step"] = step

            sname = md["shortName"]

            # print(f"message {msgcount} step {step} number {number} param {sname}")

            ldate = eccodes.codes_get_long(msg, "date")
            yyyy = math.floor(ldate / 10000)
            mm = math.floor((ldate - (yyyy * 10000)) / 100)
            dd = ldate - (yyyy * 10000) - mm * 100

            md["base_datetime"] = datetime(yyyy, mm, dd, tzinfo=timezone.utc)

            forecast_datetime = (
                datetime(yyyy, mm, dd, tzinfo=timezone.utc)
                + timedelta(minutes=60 * md["time"] / 100)
                + timedelta(minutes=60 * md["step"])
            )

            md["forecast_datetime"] = forecast_datetime

            # decode data
            # get the lats, lons, values
            # md["lats"] = eccodes.codes_get_double_array(msg, "latitudes")
            # print(lats)
            # md["lons"] = eccodes.codes_get_double_array(msg, "longitudes")
            # print(lons)
            global lats
            if lats is None:
                lats = eccodes.codes_get_double_array(msg, "latitudes")
            global lons
            if lons is None:
                lons = eccodes.codes_get_double_array(msg, "longitudes")

            md["values"] = eccodes.codes_get_double_array(msg, "values")
            # print(values)

            md["grib"] = msg  # keep grib open

            # assert sname not in messages

            messages[sname] = md

    f.close()


###########################################################################################################


# @thermofeel.timer
def calc_cossza_int(messages):
    dt = messages["2t"]["base_datetime"]
    time, step, begin, end = timestep_interval(messages)

    cossza = thermofeel.calculate_cos_solar_zenith_angle_integrated(
        lat=lats,
        lon=lons,
        y=dt.year,
        m=dt.month,
        d=dt.day,
        h=dt.hour,
        tbegin=begin,
        tend=end,
        integration_order=2,
    )

    return cossza


# @thermofeel.timer
def approximate_dsrp(messages):
    """
    In the absence of dsrp, approximate it with fdir and cossza.
    Note this introduces some amoutn of error as cossza approaches zero
    """
    fdir = messages["fdir"]["values"]
    cossza = calc_field("cossza", calc_cossza_int, messages)

    dsrp = thermofeel.approximate_dsrp(fdir, cossza)

    return dsrp


# @thermofeel.timer
def calc_heatx(messages):
    t2m = messages["2t"]["values"]
    td = messages["2d"]["values"]

    heatx = thermofeel.calculate.heat_index_adjusted(t2m=t2m, td=td)

    return heatx


# @thermofeel.timer
def calc_aptmp(messages):
    t2m = messages["2t"]["values"]

    ws = calc_field("ws", calc_ws, messages)

    aptmp = thermofeel.calculate_apparent_temperature(t2m=t2m, va=ws)

    return aptmp


# @thermofeel.timer
def calc_humidex(messages):
    t2m = messages["2t"]["values"]
    td = messages["2d"]["values"]

    humidex = thermofeel.calculate_humidex(t2m=t2m, td=td)

    return humidex


# @thermofeel.timer
def calc_rhp(messages):
    t2m = messages["2t"]["values"]
    td = messages["2d"]["values"]

    rhp = thermofeel.calculate_relative_humidity_percent(t2m=t2m, td=td)

    return rhp


# @thermofeel.timer
def calc_mrt(messages):
    time, step, begin, end = timestep_interval(messages)

    assert begin < end

    cossza = calc_field("cossza", calc_cossza_int, messages)

    # will use dsrp if available, otherwise approximate it
    dsrp = calc_field("dsrp", approximate_dsrp, messages)

    seconds_in_time_step = (end - begin) * 3600  # steps are in hours

    f = 1.0 / float(seconds_in_time_step)

    ssrd = messages["ssrd"]["values"]
    ssr = messages["ssr"]["values"]
    fdir = messages["fdir"]["values"]
    dsrp = messages["dsrp"]["values"]
    strd = messages["strd"]["values"]
    strr = messages["str"]["values"]

    mrt = thermofeel.calculate_mean_radiant_temperature(
        ssrd * f, ssr * f, dsrp * f, strd * f, fdir * f, strr * f, cossza * f
    )

    return mrt


def calc_field(name, func, messages):
    if name in results:
        return results[name]

    values = func(messages)

    field_stats(name, values)
    results[name] = values

    return values


# @thermofeel.timer
def calc_ws(messages):
    u10 = messages["10u"]["values"]
    v10 = messages["10v"]["values"]

    ws = np.sqrt(u10**2 + v10**2)

    return ws


def compute_ehPa_(rh_pc, svp):
    return svp * rh_pc * 0.01  # / 100.0


# @thermofeel.timer
def compute_ehPa(t2m, t2d):
    rh_pc = thermofeel.calculate_relative_humidity_percent(t2m, t2d)
    svp = thermofeel.calculate_saturation_vapour_pressure(t2m)
    ehPa = compute_ehPa_(rh_pc, svp)
    return ehPa


# @thermofeel.timer
def compute_utci_in_kelvin(t2m, ws, mrt, ehPa):
    utci = thermofeel.calculate_utci(t2_k=t2m, va_ms=ws, mrt_k=mrt, e_hPa=ehPa)
    return thermofeel.celsius_to_kelvin(utci)


# @thermofeel.timer
def filter_utci(t2m, va, mrt, ehPa, utci):
    e_mrt = np.subtract(mrt, t2m)

    misses = np.where(t2m >= thermofeel.celsius_to_kelvin(70))
    t = np.where(t2m <= thermofeel.celsius_to_kelvin(-70))
    misses = np.union1d(t, misses)

    t = np.where(va >= 25.0)  # 90kph
    misses = np.union1d(t, misses)

    t = np.where(ehPa > 50.0)
    misses = np.union1d(t, misses)

    t = np.where(e_mrt >= 100.0)
    misses = np.union1d(t, misses)

    t = np.where(e_mrt <= -30)
    misses = np.union1d(t, misses)

    return misses


# @thermofeel.timer
def validate_utci(utci, misses):
    utci[misses] = np.nan

    field_stats("utci", utci)

    out_of_bounds = 0
    for i in range(len(utci)):
        v = utci[i]
        if not np.isnan(v) and (v < UTCI_MIN_VALUE or v > UTCI_MAX_VALUE):
            out_of_bounds += 1
            print("UTCI [", i, "] = ", utci[i], " : lat/lon ", lats[i], lons[i])

    nmisses = len(misses)
    if nmisses > 0 or out_of_bounds > 0:
        print(f"UTCI => MISS {nmisses} out_of_bounds {out_of_bounds}")

    utci[misses] = MISSING_VALUE


# @thermofeel.timer
def calc_utci(messages):
    t2m = messages["2t"]["values"]
    t2d = messages["2d"]["values"]

    ws = calc_field("ws", calc_ws, messages)
    mrt = calc_field("mrt", calc_mrt, messages)

    ehPa = compute_ehPa(t2m, t2d)
    utci = compute_utci_in_kelvin(t2m, ws, mrt, ehPa)

    missing = filter_utci(t2m, ws, mrt, ehPa, utci)
    misses["utci"] = missing

    # validate_utci(utci, missing)

    return utci


# @thermofeel.timer
def calc_wbgt(messages):
    t2m = messages["2t"]["values"]  # Kelvin
    t2d = messages["2d"]["values"]

    ws = calc_field("ws", calc_ws, messages)
    mrt = calc_field("mrt", calc_mrt, messages)

    wbgt = thermofeel.calculate_wbgt(t2m, mrt, ws, t2d)
    wbgt = thermofeel.celsius_to_kelvin(wbgt)

    return wbgt


# @thermofeel.timer
def calc_bgt(messages):
    t2m = messages["2t"]["values"]  # Kelvin

    ws = calc_field("ws", calc_ws, messages)
    mrt = calc_field("mrt", calc_mrt, messages)

    bgt = thermofeel.calculate_bgt(t2m, mrt, ws)
    bgt = thermofeel.celsius_to_kelvin(bgt)

    return bgt


# @thermofeel.timer
def calc_wbt(messages):
    t2m = messages["2t"]["values"]
    t2m = thermofeel.kelvin_to_celcius(t2m)

    rhp = calc_field("rhp", calc_rhp, messages)

    wbt = thermofeel.calculate_relative_humidity_percent(t2m=t2m, rh=rhp)

    return wbt


# @thermofeel.timer
def calc_net(messages):
    t2m = messages["2t"]["values"]  # Kelvin
    t2d = messages["2d"]["values"]

    ws = calc_field("ws", calc_ws, messages)

    net = thermofeel.calculate_net_effective_temperature(t2m, ws, t2d)
    net = thermofeel.celsius_to_kelvin(net)

    return net


# @thermofeel.timer
def calc_windchill(messages):
    t2m = messages["2t"]["values"]

    ws = calc_field("ws", calc_ws, messages)

    windchill = thermofeel.calculate_wind_chill(t2m, ws)

    return windchill


def check_messages(msgs):
    assert "2t" in msgs
    assert "2d" in msgs
    assert "10u" in msgs
    assert "10v" in msgs
    assert "ssrd" in msgs
    assert "ssr" in msgs
    assert "fdir" in msgs
    assert "str" in msgs
    assert "strd" in msgs

    assert lats.size == lons.size

    ftime = msgs["2t"]["forecast_datetime"]

    for k, m in msgs.items():
        assert lats.size == m["values"].size
        assert ftime == m["forecast_datetime"]


# @thermofeel.timer
def output_grib(output, msg, paramid, values, missing=None):
    """Encode field in GRIB2"""
    grib = msg["grib"]
    handle = eccodes.codes_clone(grib)
    eccodes.codes_set_long(handle, "edition", 2)
    eccodes.codes_set_string(handle, "paramId", paramid)
    eccodes.codes_set_values(handle, values)
    if missing is not None:
        eccodes.codes_set_double(handle, "missingValue", missing)
    eccodes.codes_write(handle, output)
    eccodes.codes_release(handle)


def ifs_step_intervals(step):
    """Computes the time integration interval for the IFS forecasting system given a forecast output step"""
    # assert step != 0 and step is not None

    # assert step > 0
    assert step <= 360
    if step > 0:
        return step - 3
    else:
        return step
    # if step <= 144:
    #    assert step % 3 == 0
    #   return step - 3
    # else:
    #    if step <= 360:
    #        assert step % 6 == 0
    #        return step - 6


def timestep_interval(messages):
    msg = messages["2t"]
    step = msg["step"]  # end of the forecast integration
    time = msg["time"]
    ftime = int(time / 100)  # forecast time in hours
    integration_start = ifs_step_intervals(step)  # start of forecast integration step
    step_begin = ftime + integration_start
    step_end = ftime + step
    return time, step, step_begin, step_end


# @thermofeel.timer
def process_step(args, msgs, output):
    check_messages(msgs)

    # print(f"loaded {len(msgs)} parameters: {list(msgs.keys())}")
    template = msgs["2t"]

    dt = msgs["2t"]["base_datetime"]
    time, step, step_begin, step_end = timestep_interval(msgs)
    print(
        f"dt {dt.date().isoformat()} time {time} step {step} - [{step_begin},{step_end}]"
    )

    global results
    global misses

    results = {}
    misses = {}

    # Windspeed - shortName ws
    if args.ws:
        ws = calc_field("ws", calc_ws, msgs)
        output_grib(output, template, "10", ws)

    # Cosine of Solar Zenith Angle - shortName uvcossza - ECMWF product
    # TODO: 214001 only exists for GRIB1 -- but here we use it for GRIB2 (waiting for WMO)
    if args.cossza:
        cossza = calc_field("cossza", calc_cossza_int, msgs)
        output_grib(output, template, "214001", cossza)

    # Mean Radiant Temperature - shortName mrt - ECMWF product
    if args.mrt:
        mrt = calc_field("mrt", calc_mrt, msgs)
        output_grib(output, template, "261002", mrt)

    # Univeral Thermal Climate Index - shortName utci - ECMWF product
    if args.utci:
        utci = calc_field("utci", calc_utci, msgs)
        output_grib(output, template, "261001", utci, missing=MISSING_VALUE)

    # Heat Index (adjusted) - shortName heatx - ECMWF product
    if args.heatx:
        heatx = calc_field("heatx", calc_heatx, msgs)
        output_grib(output, template, "260004", heatx)

    # Wind Chill factor - shortName wcf - ECMWF product
    if args.windchill:
        windchill = calc_field("windchill", calc_windchill, msgs)
        output_grib(output, template, "260005", windchill)

    # Apparent Temperature - shortName aptmp - ECMWF product
    if args.aptmp:
        aptmp = calc_field("aptmp", calc_aptmp, msgs)
        output_grib(output, template, "260255", aptmp)

    # Relative humidity percent at 2m - shortName 2r - ECMWF product
    if args.rhp:
        rhp = calc_field("rhp", calc_rhp, msgs)
        output_grib(output, template, "260242", rhp)

    # Humidex - shortName hx - TO BE RELEASED as ECMWF product
    # TODO: 261016 is experimental GRIB code, update once WMO publishes
    if args.humidex:
        humidex = calc_field("hmdx", calc_humidex, msgs)
        output_grib(output, template, "261016", humidex)

    # Normal Effective Temperature - shortName nefft - TO BE RELEASED as ECMWF product
    # TODO: 212002 is experimental GRIB code, update once WMO publishes
    if args.net:
        net = calc_field("net", calc_net, msgs)
        output_grib(output, template, "261018", net)

    # Globe Temperature - shortName gt
    # TODO: 212003 is experimental GRIB code, update once WMO publishes
    if args.bgt:
        bgt = calc_field("bgt", calc_bgt, msgs)
        output_grib(output, template, "261015", bgt)

    # Wet-bulb potential temperature - shortName wbt - TO BE RELEASED as ECMWF product
    # TODO: 212004 is experimental GRIB code, update once WMO publishes
    if args.wbt:
        wbt = calc_field("wbt", calc_wbt, msgs)
        output_grib(output, template, "261022", wbt)

    # Wet Bulb Globe Temperature - shortName wbgt - TO BE RELEASED as ECMWF product
    if args.wbgt:  #
        wbgt = calc_field("wbgt", calc_wbgt, msgs)
        output_grib(output, template, "261014", wbgt)

    # effective temperature 261017
    # standard effective temperature 261019

    return step


def command_line_options():
    parser = argparse.ArgumentParser()

    parser.add_argument("input", help="input file with GRIB messages")
    parser.add_argument("output", help="output file with GRIB messages")

    parser.add_argument(
        "--ws", help="compute wind speed from components", action="store_true"
    )
    parser.add_argument(
        "--cossza",
        help="compute Cosine of Solar Zenith Angle (cossza)",
        action="store_true",
    )
    parser.add_argument("--mrt", help="compute mrt", action="store_true")
    parser.add_argument(
        "--utci",
        help="compute UTCI Universal Thermal Climate Index",
        action="store_true",
    )
    parser.add_argument(
        "--heatx", help="compute Heat Index (adjusted)", action="store_true"
    )
    parser.add_argument(
        "--windchill", help="compute Windchill factor", action="store_true"
    )
    parser.add_argument(
        "--aptmp", help="compute Apparent Temperature", action="store_true"
    )
    parser.add_argument(
        "--rhp", help="compute relative humidity percent", action="store_true"
    )

    parser.add_argument("--humidex", help="compute humidex", action="store_true")
    parser.add_argument(
        "--net", help="compute net effective temperature", action="store_true"
    )

    # TODO: these outputs are not yet in WMO GRIB2 recognised parameters
    parser.add_argument(
        "--wbgt", help="compute Wet Bulb Globe Temperature", action="store_true"
    )
    parser.add_argument("--bgt", help="compute  Globe Temperature", action="store_true")
    parser.add_argument(
        "--wbt", help="compute Wet Bulb Temperature", action="store_true"
    )

    args = parser.parse_args()

    return args


def main():
    args = command_line_options()

    print(f"Thermofeel version: {thermofeel.__version__}")
    print(f"Python version: {sys.version}")
    print(f"Numpy version: {np.version.version}")
    # np.show_config()

    output = open(args.output, "wb")

    steps = []

    print("----------------------------------------")
    for msgs in decode_grib(args.input):
        step = process_step(args, msgs, output)
        steps.append(step)
        print("----------------------------------------")

    print(f"\nProcessed steps: {steps}\n")

    print("Performance summary:")
    print("--------------------")
    for func, stats in thermofeel.func_timers.items():
        assert stats["calls"] > 0
        average = stats["elapsed"] / stats["calls"]
        print(func, "->", stats, f"average {average:0.6f} s")


if __name__ == "__main__":
    sys.exit(main())
